/*
 Copyright (C) 2010-2016 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <unistd.h>
#include "tfio.h"

#define DEFAULT_IOfkt (IOfkt_CanWrite| IOfkt_CanSave | IOfkt_CanSaveAs)

#define TCSelf TFIOmem
typedef struct TFIOmem {
  AlgTFIO
  int64_t Offset;
  int64_t mem_size;
  char *mem;
  bool Changed;
} TFIOmem;
#undef TCSelf

static int64_t fio_GetSizeAbs(TFIOmem *self) {
	if (!self->mem) return 0;
	return self->mem_size;
}

static int64_t fio_GetSize(TFIOmem *self) {
	return fio_GetSizeAbs(self) - self->Offset;
}

static bool fio_SetOffset(TFIOmem *self, int64_t Offset) {
  int64_t max;

	if (Offset < 0) Offset = 0;
	max = fio_GetSizeAbs(self);
	if (Offset > max) Offset = max;
	if (self->Offset != Offset) {
	    self->Offset = Offset;
	    if (self->OnSetOffset) self->OnSetOffset(self, self->Offset);
	}
	return true;
}

static int64_t fio_GetOffset(TFIOmem *self) {
	return self->Offset;
}

static char *fio_GetDataP(TFIOmem *self, int64_t Offset, int MaxChr) {
	if (!self->mem) return NULL;
	Offset += self->Offset;
	if ((Offset < 0) || (Offset >= fio_GetSizeAbs(self))) return NULL;
	return self->mem + Offset;
}

static bool fio_OverWrite(TFIOmem *self, int64_t Offset, const char *s, int size) {
  int64_t max;
  char *pc;

	if (!s || (size <= 0)) return true;
	Offset += self->Offset;
	max = fio_GetSizeAbs(self);
	if (max > 0x7fffffff) return false;
	if ((Offset < 0) || (Offset >= max)) return false;
	if ((Offset + size) >= max) size = max - Offset;
	if (size <= 0) return true;
	pc = self->mem + Offset;
	pnmove(pc, s, size);
	self->Changed = true;
	return True;
}

static bool fio_Close(TFIOmem *self) {
	ppfree(&self->mem);
	self->mem_size = 0;
	self->Offset = 0;
	self->Changed = false;
	return true;
}

static bool fio_Open(TFIOmem *self) {
//	TFIOlp_Close(self);
	return true;
}

static bool fio_Save(TFIOmem *self, const char *FileName) {
  FileHandle fd;
  size_t idx, bufsize, retsize;


	if (!self->mem_size || !FileName || 
	    !file_open_wr_ae(FileName, &fd, 0, EMSG_ERR1)) return false;

	if (FileName) ppcchange_cmp(&self->FileName, FileName);

	idx = 0;
	bufsize = self->mem_size;

	while ((bufsize > 0) && 
	       file_write_ae(fd, self->mem + idx, bufsize, &retsize, FileName, EMSG_ERR1)) {
	    idx += retsize;
	    bufsize -= retsize;
	}
	file_close(fd, NULL);

	if (self->OnWriteEnd) self->OnWriteEnd(self);
	return bufsize == 0;
}

static bool fio_Load(TFIOmem *self, const char *FileName) {
  Tstat64 *st;
  FileHandle fd;
  size_t idx, bufsize, retsize;

	if (!FileName) return false;
	st = file_stat_fn(FileName, fsFollowLink);
	if (!st || (st->size > 0x7fffffff)) return false;
	FIO_SetMemSize((TFIO*)self, st->size);
	xfree(st);

	if (!file_open_rd_ae(FileName, &fd, EMSG_ERR1)) return false;
	if (FileName) ppcchange_cmp(&self->FileName, FileName);

	idx = 0;
	bufsize = self->mem_size;

	while ((bufsize > 0) && 
	       file_read_ae(fd, self->mem + idx, bufsize, &retsize, FileName, EMSG_ERR1)) {
	    idx += retsize;
	    bufsize -= retsize;
	}
	file_close(fd, NULL);

	if (self->OnReadEnd) self->OnReadEnd(self);
	return bufsize == 0;
}

static bool fio_SetMemSize(TFIOmem *self, int64_t size) {
  int64_t tsize;

	tsize = self->mem_size;
	if (size <= 0) {
	    fio_Close(self);
	} else if (size > 0x40000000) {
	    size = 0x40000000;
	}
	if ((size > 0) && (self->mem_size != size)) {
	    if (self->mem_size > 0) {
		self->mem = xrecalloc(self->mem, size, self->mem_size);
	    } else {
		self->mem = xcalloc(size);
	    }
	    self->mem_size = size;
	}
	return (self->mem_size != tsize);
}

/*____________________________ Destroy __________________________________*/
static bool TFIOmem__Destroy(TFIOmem *self) {
	fio_Close(self);
	return TFIO__Destroy((TFIO*)self);
}

/*____________________________ Create ___________________________________*/
TFIO *TFIO_Create_MEM() {
  TFIOmem *self;

	self = (TFIOmem*)TFIO_Create(NULL, sizeof(TFIOmem), FIOartMEM);
	self->destroy = (TDESTROY)TFIOmem__Destroy;

	self->Open = fio_Open;
	self->GetDataP = fio_GetDataP;
	self->GetSize = fio_GetSize;
	self->GetSizeAbs = fio_GetSizeAbs;
	self->SetOffset = fio_SetOffset;
	self->GetOffset = fio_GetOffset;
	self->SetMemSize = fio_SetMemSize;
	self->OverWrite = fio_OverWrite;
	self->Save = fio_Save;
	self->Load = fio_Load;

/*
	self->lb = TLineBuf_Create();
	self->lb->wr_memlook = True;

//	self->ReOpen = TFIOmem_ReOpen;
	self->Save = TFIOmem_Save;
//	self->SaveAs = TFIOmem_SaveAs;
	self->GetChanged = GetChanged;
	self->Insert = TFIOmem_Insert;
	self->Delete = TFIOmem_Delete;
*/
	self->IOfkt = DEFAULT_IOfkt;

	return (TFIO*)self;
}
